@@ -1,10 +1,11 @@
-from mindsdb.libs.constants.mindsdb import *
-from mindsdb.config import *
-from mindsdb.libs.helpers.general_helpers import disable_ludwig_output
-
 from dateutil.parser import parse as parse_datetime
 import os, sys
 import shutil
+import subprocess
+
+from mindsdb.libs.constants.mindsdb import *
+from mindsdb.config import *
+from mindsdb.libs.helpers.general_helpers import disable_console_output, get_tensorflow_colname
 
 from tensorflow.python.client import device_lib
 from ludwig.api import LudwigModel
@@ -17,6 +18,21 @@ from imageio import imread
 class LudwigBackend():
 
     def __init__(self, transaction):
+        try:
+            subprocess.call(['python3','-m','spacy','download','en_core_web_sm'])
+        except:
+            try:
+                subprocess.call(['python','-m','spacy','download','en_core_web_sm'])
+            except:
+                print('Can\'t download spacy vocabulary, ludwig backend may fail when processing text input')
+        try:
+            subprocess.call(['python3','-m','spacy','download','en'])
+        except:
+            try:
+                subprocess.call(['python','-m','spacy','download','en'])
+            except:
+                print('Can\'t download spacy vocabulary, ludwig backend may fail when processing text input')
+
         self.transaction = transaction
 
     def _translate_df_to_timeseries_format(self, df, model_definition, timeseries_cols, mode='predict'):
@@ -59,120 +75,68 @@ class LudwigBackend():
 
         nr_ele = len(df[timeseries_col_name])
 
-        if self.transaction.lmd['window_size_seconds'] is not None:
-            window_size_seconds = self.transaction.lmd['window_size_seconds']
-            i = 0
-            while i < nr_ele:
-                current_window = 0
-                new_row = {}
+        i = 0
+        while i < nr_ele:
+            new_row = {}
 
-                timeseries_row = [df[timeseries_col_name][i]]
+            timeseries_row = [df[timeseries_col_name][i]]
+
+            for col in other_col_names:
+                new_row[col] = [df[col][i]]
+            for col in previous_predict_col_names:
+                new_row[col] = []
+            for col in predict_col_names:
+                new_row[col] = df[col][i]
+            for col in self.transaction.lmd['model_group_by']:
+                new_row[col] = df[col][i]
+
+            inverted_index_range = list(range(i))
+            inverted_index_range.reverse()
+            ii = 0
+            for ii in inverted_index_range:
+                if (i - ii) > self.transaction.lmd['window_size']:
+                    break
+                timeseries_row.append(df[timeseries_col_name][ii])
 
                 for col in other_col_names:
-                    new_row[col] = [df[col][i]]
+                    new_row[col].append(df[col][ii])
                 for col in previous_predict_col_names:
-                    new_row[col] = []
-                for col in predict_col_names:
-                    new_row[col] = df[col][i]
-                for col in self.transaction.lmd['model_group_by']:
-                    new_row[col] = df[col][i]
-
-                inverted_index_range = list(range(i))
-                inverted_index_range.reverse()
-                ii = 0
-                for ii in inverted_index_range:
-                    if window_size_seconds < current_window + (timeseries_row[-1] - df[timeseries_col_name][ii]):
-                        break
-                    current_window += (timeseries_row[-1] - df[timeseries_col_name][ii])
-                    timeseries_row.append(df[timeseries_col_name][ii])
-
-                    for col in other_col_names:
-                        new_row[col].append(df[col][ii])
-                    for col in previous_predict_col_names:
+                    try:
+                        new_row[col].append(df[col.replace('previous_', '')][ii])
+                    except:
                         try:
-                            new_row[col].append(df[col.replace('previous_', '')][ii])
+                            new_row[col].append(df[col][ii])
                         except:
-                            try:
-                                new_row[col].append(df[col][ii])
-                            except:
-                                self.transaction.log.warning('Missing previous predicted values for output column: {}, these should be included in your input under the name: {}'.format(col.replace('previous_', ''), col))
+                            self.transaction.log.warning('Missing previous predicted values for output column: {}, these should be included in your input under the name: {}'.format(col.replace('previous_', ''), col))
 
-                if mode == 'train':
-                    i = max(i + 1, (i + round((i - ii)/2)))
-                else:
-                    i = i + 1
+            if mode == 'train':
+                i = max(i + 1, (i + round((i - ii)/2)))
+            else:
+                i = i + 1
 
-                new_row[timeseries_col_name] = timeseries_row
+            new_row[timeseries_col_name] = timeseries_row
 
-                for col in new_row:
-                    if col not in predict_col_names and col not in self.transaction.lmd['model_group_by']:
-                        new_row[col].reverse()
-                    new_cols[col].append(new_row[col])
-        else:
-            window_size_samples = self.transaction.lmd['window_size_samples']
-            i = 0
-            while i < nr_ele:
-                new_row = {}
-
-                timeseries_row = [df[timeseries_col_name][i]]
-
-                for col in other_col_names:
-                    new_row[col] = [df[col][i]]
-                for col in previous_predict_col_names:
-                    new_row[col] = []
-                for col in predict_col_names:
-                    new_row[col] = df[col][i]
-                for col in self.transaction.lmd['model_group_by']:
-                    new_row[col] = df[col][i]
-
-                inverted_index_range = list(range(i))
-                inverted_index_range.reverse()
-                ii = 0
-                for ii in inverted_index_range:
-                    if (i - ii) > window_size_samples:
-                        break
-                    timeseries_row.append(df[timeseries_col_name][ii])
-
-                    for col in other_col_names:
-                        new_row[col].append(df[col][ii])
-                    for col in previous_predict_col_names:
-                        try:
-                            new_row[col].append(df[col.replace('previous_', '')][ii])
-                        except:
-                            try:
-                                new_row[col].append(df[col][ii])
-                            except:
-                                self.transaction.log.warning('Missing previous predicted values for output column: {}, these should be included in your input under the name: {}'.format(col.replace('previous_', ''), col))
-
-                if mode == 'train':
-                    i = max(i + 1, (i + round((i - ii)/2)))
-                else:
-                    i = i + 1
-
-                new_row[timeseries_col_name] = timeseries_row
-
-                for col in new_row:
-                    if col not in predict_col_names and col not in self.transaction.lmd['model_group_by']:
-                        new_row[col].reverse()
-                    new_cols[col].append(new_row[col])
+            for col in new_row:
+                if col not in predict_col_names and col not in self.transaction.lmd['model_group_by']:
+                    new_row[col].reverse()
+                new_cols[col].append(new_row[col])
 
         new_df = pd.DataFrame(data=new_cols)
         df = new_df
         return df, model_definition
 
     def _create_ludwig_dataframe(self, mode):
+        has_heavy_data = False
+        col_map = {}
+
         if mode == 'train':
-            indexes = self.transaction.input_data.train_indexes[KEY_NO_GROUP_BY]
-            columns = [[col, col_ind] for col_ind, col in enumerate(self.transaction.lmd['columns'])]
+            df = self.transaction.input_data.train_df
         elif mode == 'predict':
-            indexes = self.transaction.input_data.all_indexes[KEY_NO_GROUP_BY]
-            columns = [[col, col_ind] for col_ind, col in enumerate(self.transaction.lmd['columns']) if col not in self.transaction.lmd['predict_columns']]
+            df = self.transaction.input_data.data_frame
         elif mode == 'validate':
-            indexes = self.transaction.input_data.validation_indexes[KEY_NO_GROUP_BY]
-            columns = [[col, col_ind] for col_ind, col in enumerate(self.transaction.lmd['columns'])]  #if col not in self.transaction.lmd['predict_columns']]
+            df = self.transaction.input_data.validation_df
         elif mode == 'test':
-            indexes = self.transaction.input_data.test_indexes[KEY_NO_GROUP_BY]
-            columns = [[col, col_ind] for col_ind, col in enumerate(self.transaction.lmd['columns'])] #if col not in self.transaction.lmd['predict_columns']]
+            df = self.transaction.input_data.test_df
         else:
             raise Exception(f'Unknown mode specified: "{mode}"')
         model_definition = {'input_features': [], 'output_features': []}
@@ -183,15 +147,15 @@ class LudwigBackend():
         else:
             timeseries_cols = list(map(lambda x: x[0], self.transaction.lmd['model_order_by']))
 
-        for ele in columns:
-            col = ele[0]
-            col_ind = ele[1]
+        for col in df.columns:
+            tf_col = get_tensorflow_colname(col)
+            col_map[tf_col] = col
 
             # Handle malformed columns
-            if col in self.transaction.lmd['malformed_columns']['names']:
+            if col in self.transaction.lmd['columns_to_ignore']:
                 continue
 
-            data[col] = []
+            data[tf_col] = []
 
             col_stats = self.transaction.lmd['column_stats'][col]
             data_subtype = col_stats['data_subtype']
@@ -232,6 +196,7 @@ class LudwigBackend():
                 ludwig_dtype = 'category'
 
             elif data_subtype in (DATA_SUBTYPES.IMAGE):
+                has_heavy_data = True
                 ludwig_dtype = 'image'
                 encoder = 'stacked_cnn'
                 in_memory = True
@@ -249,50 +214,50 @@ class LudwigBackend():
 
             custom_logic_continue = False
 
-            for row_ind in indexes:
+            for index, row in df.iterrows():
                 if ludwig_dtype == 'order_by_col':
-                    ts_data_point = self.transaction.input_data.data_array[row_ind][col_ind]
+                    ts_data_point = row[col]
 
                     try:
                         ts_data_point = float(ts_data_point)
                     except:
                         ts_data_point = parse_datetime(ts_data_point).timestamp()
-                    data[col].append(ts_data_point)
+                    data[tf_col].append(ts_data_point)
 
                 elif ludwig_dtype == 'sequence':
-                    arr_str = self.transaction.input_data.data_array[row_ind][col_ind]
+                    arr_str = row[col]
                     if arr_str is not None:
                         arr = list(map(float,arr_str.rstrip(']').lstrip('[').split(self.transaction.lmd['column_stats'][col]['separator'])))
                     else:
                         arr = ''
-                    data[col].append(arr)
+                    data[tf_col].append(arr)
 
                 # Date isn't supported yet, so we hack around it
                 elif ludwig_dtype == 'date':
                     if col in data:
                         data.pop(col)
-                        data[col + '_year'] = []
-                        data[col + '_month'] = []
-                        data[col + '_day'] = []
+                        data[tf_col + '_year'] = []
+                        data[tf_col + '_month'] = []
+                        data[tf_col + '_day'] = []
 
                         model_definition['input_features'].append({
                             'name': col + '_year'
-                            ,'type': 'numerical'
+                            ,'type': 'category'
                         })
                         model_definition['input_features'].append({
                             'name': col + '_month'
-                            ,'type': 'numerical'
+                            ,'type': 'category'
                         })
                         model_definition['input_features'].append({
                             'name': col + '_day'
                             ,'type': 'numerical'
                         })
 
-                    date = parse_datetime(self.transaction.input_data.data_array[row_ind][col_ind])
+                    date = parse_datetime(row[col])
 
-                    data[col + '_year'].append(date.year)
-                    data[col + '_month'].append(date.month)
-                    data[col + '_day'].append(date.day)
+                    data[tf_col + '_year'].append(date.year)
+                    data[tf_col + '_month'].append(date.month)
+                    data[tf_col + '_day'].append(date.day)
 
                     custom_logic_continue = True
 
@@ -303,34 +268,39 @@ class LudwigBackend():
                         timeseries_cols.append(col + '_year')
 
                 elif data_subtype in (DATA_SUBTYPES.TIMESTAMP):
-                    if self.transaction.input_data.data_array[row_ind][col_ind] is None:
+                    if row[col] is None:
                         unix_ts = 0
                     else:
-                        unix_ts = parse_datetime(self.transaction.input_data.data_array[row_ind][col_ind]).timestamp()
+                        unix_ts = parse_datetime(row[col]).timestamp()
 
-                    data[col].append(unix_ts)
+                    data[tf_col].append(unix_ts)
 
                 elif data_subtype in (DATA_SUBTYPES.FLOAT):
-                    if type(self.transaction.input_data.data_array[row_ind][col_ind]) == str:
-                        data[col].append(float(str(self.transaction.input_data.data_array[row_ind][col_ind]).replace(',','.')))
+                    if type(row[col]) == str:
+                        data[tf_col].append(float(str(row[col]).replace(',','.')))
                     else:
-                        data[col].append(self.transaction.input_data.data_array[row_ind][col_ind])
+                        data[tf_col].append(row[col])
 
                 elif data_subtype in (DATA_SUBTYPES.INT):
-                    if type(self.transaction.input_data.data_array[row_ind][col_ind]) == str:
-                        data[col].append(round(float(str(self.transaction.input_data.data_array[row_ind][col_ind]).replace(',','.'))))
+                    if type(row[col]) == str:
+                        data[tf_col].append(round(float(str(row[col]).replace(',','.'))))
                     else:
-                        data[col].append(self.transaction.input_data.data_array[row_ind][col_ind])
+                        data[tf_col].append(row[col])
 
+                elif data_subtype in (DATA_SUBTYPES.IMAGE):
+                    if os.path.isabs(row[col]):
+                        data[tf_col].append(row[col])
+                    else:
+                        data[tf_col].append(os.path.join(os.getcwd(), row[col]))
                 else:
-                    data[col].append(self.transaction.input_data.data_array[row_ind][col_ind])
+                    data[tf_col].append(row[col])
 
             if custom_logic_continue:
                 continue
 
             if col not in self.transaction.lmd['predict_columns']:
                 input_def = {
-                    'name': col
+                    'name': tf_col
                     ,'type': ludwig_dtype
                 }
                 if encoder is not None:
@@ -358,7 +328,7 @@ class LudwigBackend():
                 model_definition['input_features'].append(input_def)
             else:
                 output_def = {
-                    'name': col
+                    'name': tf_col
                     ,'type': ludwig_dtype
                 }
                 model_definition['output_features'].append(output_def)
@@ -367,9 +337,9 @@ class LudwigBackend():
         if len(timeseries_cols) > 0:
             df.sort_values(timeseries_cols)
 
-        return df, model_definition, timeseries_cols
+        return df, model_definition, timeseries_cols, has_heavy_data, col_map
 
-    def get_model_dir(self):
+    def _get_model_dir(self):
         model_dir = None
         for thing in os.listdir(self.transaction.lmd['ludwig_data']['ludwig_save_path']):
             if 'api_experiment' in thing:
@@ -378,7 +348,7 @@ class LudwigBackend():
             model_dir = os.path.join(self.transaction.lmd['ludwig_data']['ludwig_save_path'],'model')
         return model_dir
 
-    def get_useable_gpus(self):
+    def _get_useable_gpus(self):
         if self.transaction.lmd['use_gpu'] == False:
             return []
         local_device_protos = device_lib.list_local_devices()
@@ -391,12 +361,12 @@ class LudwigBackend():
             return gpu_indices
 
     def train(self):
-        training_dataframe, model_definition, timeseries_cols = self._create_ludwig_dataframe('train')
+        training_dataframe, model_definition, timeseries_cols, has_heavy_data, self.transaction.lmd['ludwig_tf_self_col_map'] = self._create_ludwig_dataframe('train')
 
         if len(timeseries_cols) > 0:
             training_dataframe, model_definition =  self._translate_df_to_timeseries_format(training_dataframe, model_definition, timeseries_cols, 'train')
 
-        with disable_ludwig_output(True):
+        with disable_console_output(True):
             # <---- Ludwig currently broken, since mode can't be initialized without train_set_metadata and train_set_metadata can't be obtained without running train... see this issue for any updates on the matter: https://github.com/uber/ludwig/issues/295
             #model.initialize_model(train_set_metadata={})
             #train_stats = model.train_online(data_df=training_dataframe) # ??Where to add model_name?? ----> model_name=self.transaction.lmd['name']
@@ -415,12 +385,30 @@ class LudwigBackend():
                     merged_model_definition['output_features']),
                     merged_model_definition['preprocessing']
                 )
-                model.initialize_model(train_set_metadata=train_set_metadata, gpus=self.get_useable_gpus())
-
-                train_stats = model.train(data_df=training_dataframe, model_name=self.transaction.lmd['name'], skip_save_model=ludwig_save_is_working, skip_save_progress=True, gpus=self.get_useable_gpus())
+                model.initialize_model(train_set_metadata=train_set_metadata, gpus=self._get_useable_gpus())
             else:
-                model = LudwigModel.load(model_dir=self.get_model_dir())
-                train_stats = model.train(data_df=training_dataframe, model_name=self.transaction.lmd['name'], skip_save_model=ludwig_save_is_working, skip_save_progress=True, gpus=self.get_useable_gpus())
+                model = LudwigModel.load(model_dir=self._get_model_dir())
+
+
+            split_by = int(20 * pow(10,6))
+            if has_heavy_data:
+                split_by = 40
+            df_len = len(training_dataframe[training_dataframe.columns[0]])
+            if df_len > split_by:
+                i = 0
+                while i < df_len:
+                    end = i + split_by
+                    self.transaction.log.info(f'Training with batch from index {i} to index {end}')
+                    training_sample = training_dataframe.iloc[i:end]
+                    training_sample = training_sample.reset_index()
+
+                    if len(training_sample) < 1:
+                        continue
+
+                    train_stats = model.train(data_df=training_sample, model_name=self.transaction.lmd['name'], skip_save_model=ludwig_save_is_working, skip_save_progress=True, gpus=self._get_useable_gpus())
+                    i = end
+            else:
+                train_stats = model.train(data_df=training_dataframe, model_name=self.transaction.lmd['name'], skip_save_model=ludwig_save_is_working, skip_save_progress=True, gpus=self._get_useable_gpus())
 
             for k in train_stats['train']:
                 if k not in self.transaction.lmd['model_accuracy']['train']:
@@ -452,26 +440,31 @@ class LudwigBackend():
         self.transaction.hmd['ludwig_data'] = {'model_definition': model_definition}
 
     def predict(self, mode='predict', ignore_columns=[]):
-        predict_dataframe, model_definition, timeseries_cols = self._create_ludwig_dataframe(mode)
+        predict_dataframe, model_definition, timeseries_cols, has_heavy_data, _ = self._create_ludwig_dataframe(mode)
         model_definition = self.transaction.hmd['ludwig_data']['model_definition']
 
         if len(timeseries_cols) > 0:
             predict_dataframe, model_definition =  self._translate_df_to_timeseries_format(predict_dataframe, model_definition, timeseries_cols)
 
         for ignore_col in ignore_columns:
+            for tf_col in self.transaction.lmd['ludwig_tf_self_col_map']:
+                if ignore_col == self.transaction.lmd['ludwig_tf_self_col_map'][tf_col]:
+                    ignore_col = tf_col
             try:
                 predict_dataframe[ignore_col] = [None] * len(predict_dataframe[ignore_col])
             except:
                 for date_appendage in ['_year', '_month','_day']:
                     predict_dataframe[ignore_col + date_appendage] = [None] * len(predict_dataframe[ignore_col + date_appendage])
 
-        with disable_ludwig_output(True):
-            model_dir = self.get_model_dir()
+        with disable_console_output(True):
+            model_dir = self._get_model_dir()
             model = LudwigModel.load(model_dir=model_dir)
-            predictions = model.predict(data_df=predict_dataframe, gpus=self.get_useable_gpus())
+            predictions = model.predict(data_df=predict_dataframe, gpus=self._get_useable_gpus())
 
         for col_name in predictions:
             col_name_normalized = col_name.replace('_predictions', '')
+            if col_name_normalized in self.transaction.lmd['ludwig_tf_self_col_map']:
+                col_name_normalized = self.transaction.lmd['ludwig_tf_self_col_map'][col_name_normalized]
             predictions = predictions.rename(columns = {col_name: col_name_normalized})
 
         return predictions
