@@ -3,21 +3,23 @@ from mindsdb.libs.constants.mindsdb import *
 from mindsdb.libs.data_types.mindsdb_logger import log
 from mindsdb.libs.data_types.transaction_output_row import TransactionOutputRow
 
+
 class TrainTransactionOutputData():
     def __init__(self):
-        self.data_array = None
+        self.data_frame = None
         self.columns = None
 
 
 class PredictTransactionOutputData():
-    def __init__(self, transaction = None, data = {}, evaluations = {}, ):
+    def __init__(self, transaction, data, evaluations):
         self.data = data
         self.evaluations = evaluations
         self.transaction = transaction
+        self.input_confidence_arr = None
 
     def __iter__(self):
-        for i, cell in enumerate(self.data[self.transaction.lmd['columns'][0]]):
-            yield TransactionOutputRow(self, i).as_dict()
+        for i, value in enumerate(self.data[self.transaction.lmd['columns'][0]]):
+            yield TransactionOutputRow(self, i)
 
     def __getitem__(self, item):
         return TransactionOutputRow(self, item)
